/*=======================================================================*/
/*
 * Parallel Recursive Spectral Bisection -
 * Based on the sequential version provided by Horst Simon
 * -ravi
 */
 /*=======================================================================*/

#include <stdio.h>
#include "parti.h"

typedef struct {
  double data;
  int perm;
} crib_t;

/*======================================================*/

int compar(x,y)
crib_t *x,*y;
{
  if(x->data > y->data) return(1);
  if(x->data < y->data) return(-1);
  return(0);
}

/*======================================================*/

/*
main(argc,argv)
int argc;
char *argv[];
{
  int i,ndata;
  double data[100],wrk[200];
  int p[100];
  FILE *fp;
  int distribution[128],mydata,maxdata;


  for(i=0;i<PARTI_numnodes;i++){
    distribution[i] = atoi(argv[i+1]);
  }
  mydata = distribution[PARTI_mynode];
  maxdata = mydata;
  for(i=0;i<PARTI_numnodes;i++){
    if(maxdata<distribution[i]) maxdata = distribution[i];
  }

  read_data("data",wrk,distribution,data);

  crib_sort(data,mydata,p);
  print_data(distribution,p);
}
*/

/*======================================================*/
crib_sort(d,n,p)
double *d;
int *p;
int n;
{
  int *perm,count,i,j,totaln,distrib[128];
  double *data;

  for(i=0;i<PARTI_numnodes;i++) 
    distrib[i] = 0;
  distrib[PARTI_mynode] = n;
#ifdef MYDEBUG
  printf("\ncrib_sort: n=%d", n);
#endif
  PARTI_gcol(&n,sizeof(int),distrib,128*sizeof(int),&count);
#ifdef MYDEBUG
  printf("\ncrib_sort: d0=%d, d1=%d cnt=%d", *distrib, *(distrib+1), count);
#endif

  if(PARTI_mynode==0){

    totaln = 0;
    for(i=0;i<PARTI_numnodes;i++) totaln += distrib[i];
    data = (double *) malloc(sizeof(double)*totaln);
    perm = (int *) malloc(sizeof(int)*totaln);
    if (data == NULL || perm == NULL) 
      printf("Not enough memory for sorting:Node %d \n", PARTI_mynode);

    count = 0;
    for(i=0;i<n;i++){
      data[count++] = d[i];
    }

    for(i=1;i<PARTI_numnodes;i++){
      PARTI_csend(101,NULL,0,i,PARTI_pid);
      PARTI_crecv(102,&data[count],sizeof(double)*distrib[i],0,PARTI_pid);
      count += distrib[i];
    }

    local_sort(data,totaln,perm);

    count = 0;
    for(i=0;i<n;i++) p[i] = perm[count++];
    for(i=1;i<PARTI_numnodes;i++){
      PARTI_csend(103,&perm[count],sizeof(int)*distrib[i],i,PARTI_pid);
      count += distrib[i];
    }

    free(data);
    free(perm);
  }else{
    PARTI_crecv(101,NULL,0,0,PARTI_pid);
    PARTI_csend(102,d,sizeof(double)*n,0,PARTI_pid);
    PARTI_crecv(103,p,sizeof(int)*n,0,PARTI_pid);
  }
}
/*======================================================*/

local_sort(d,n,perm)
double *d;
int n;
int *perm;
{
  crib_t *data;
  int i;

  data = (crib_t *) malloc(sizeof(crib_t)*n);
  if (data == NULL ) 
      printf("Not enough memory for sorting:Node %d \n", PARTI_mynode);
  for(i=0;i<n;i++){
    data[i].perm = i;
    data[i].data = d[i];
  }

  qsort(data,n,sizeof(crib_t),compar);

  for(i=0;i<n;i++){
    d[i] = data[i].data;
    perm[i] = data[i].perm;
  }
  free(data);
}
/*======================================================*/

print_data(distribution,data)
int *distribution;
int *data;
{
  int int_tmp,count,i,j,mydata;
  int buffer[1000];

  mydata = distribution[PARTI_mynode];
  if(PARTI_mynode==0){

    /* print data from node #0 */
    count = 0;
    printf(" 0 : ");
    for(i=0;i<mydata;i++){
      printf("%4d ",data[i]);
      count++;
      if(count==15){
        printf("\n");
        printf("     ");
        count = 0;
      }
    }
    printf("\n");

    /* get and print data from nodes other from #0 */
    for(j=1;j<PARTI_numnodes;j++){
      PARTI_csend(101,NULL,0,j,PARTI_pid);
      PARTI_crecv(102,buffer,sizeof(int)*distribution[j],j,PARTI_pid);
      printf("%2d : ",j);
      count = 0;
      for(i=0;i<distribution[j];i++){
        printf("%4d ",buffer[i]);
        count++;
        if(count==15){
          printf("\n");
          printf("     ");
          count = 0;
        }
      }
      printf("\n");
    }
  }else{

    PARTI_crecv(101,NULL,0,0,PARTI_pid);
    PARTI_csend(102,data,mydata*sizeof(int),0,PARTI_pid);
  }
}
/*======================================================*/

read_data(fname,wrk,distribution,data)
char *fname;
int *distribution;
double *data,*wrk;
{
  int i,j,mydata;
  FILE *fp;

  mydata = distribution[PARTI_mynode];
  if(PARTI_mynode==0){

    fp = fopen(fname,"r");
    if(fp==NULL){
      fprintf(stderr,"Error: cannot open %s for input\n",fname);
      exit();
    }

    /* read data edstined for node 0 */
    for(i=0;i<mydata;i++){
      fscanf(fp,"%lf",&data[i]);
    }

    /* read and send data to other nodes */
    for(i=1;i<PARTI_numnodes;i++){
      for(j=0;j<distribution[i];j++){
        fscanf(fp,"%lf",&wrk[j]);
      }
      PARTI_csend(1,wrk,sizeof(double)*distribution[i],i,PARTI_pid);
    }
    fclose(fp);

  }else{
    PARTI_crecv(1,data,sizeof(double)*mydata,0,PARTI_pid);
  }
}
/*======================================================*/
